
package fast.maps;

import java.util.Random;

public class Cuckoo3<V> {
	static private final int EMPTY = Integer.MIN_VALUE;
	private static final int PRIME1 = 0xbe1f14b1;
	private static final int PRIME2 = 0x000196d3;
	private static final int PRIME3 = 0xced1c241;

	public int size;

	int[] keyTable;
	V[] valueTable;

	private float loadFactor;
	private int hashShift, mask, capacity, threshold;
	private int pushIterations;

	public Cuckoo3 (int initialCapacity, float loadFactor) {
		if (initialCapacity > 1 << 30) throw new IllegalArgumentException("initialCapacity is too large.");
		if (initialCapacity < 0) throw new IllegalArgumentException("initialCapacity must be greater than zero.");
		if (loadFactor <= 0) throw new IllegalArgumentException("initialCapacity must be greater than zero.");
		capacity = nextPowerOfTwo((int)(initialCapacity * (1 / loadFactor)));
		hashShift = 31 - Integer.numberOfTrailingZeros(capacity);
		this.loadFactor = loadFactor;
		threshold = (int)(capacity * loadFactor);
		pushIterations = Math.max(32, (int)Math.sqrt(capacity) / 2);
		keyTable = new int[capacity];
		for (int i = keyTable.length - 1; i >= 0; i--)
			keyTable[i] = EMPTY;
		valueTable = (V[])new Object[capacity];
		mask = capacity - 1;
	}

	public static void main (String[] args) {
		int c = 5000;
		Random r = new Random();
		for (int i = 0; i < 1500000; i++) {
			Cuckoo3 m = new Cuckoo3(c, 0.9f);
			for (int ii = 0; ii < c; ii++) {
				m.put(r.nextInt(), "moo" + ii);
				// m.put(i, "moo" + i);
			}
			// System.out.println(m.size);
			System.out.println(i);
			m.clear();
		}
	}

	public void put (int key, V value) {
		int[] keyTable = this.keyTable;
		int mask = this.mask;

		// Check for existing keys.
		int index1 = key & mask;
		int existingKey1 = keyTable[index1];
		if (existingKey1 == key) {
			valueTable[index1] = value;
			return;
		}

		int index2 = hash2(key) & mask;
		int existingKey2 = keyTable[index2];
		if (existingKey2 == key) {
			valueTable[index2] = value;
			return;
		}

		int index3 = hash3(key) & mask;
		int existingKey3 = keyTable[index3];
		if (existingKey3 == key) {
			valueTable[index3] = value;
			return;
		}

		// Check for empty buckets.
		if (existingKey1 == EMPTY) {
			keyTable[index1] = key;
			valueTable[index1] = value;
			if (size++ >= threshold) resize(capacity << 1);
			return;
		}

		if (existingKey2 == EMPTY) {
			keyTable[index2] = key;
			valueTable[index2] = value;
			if (size++ >= threshold) resize(capacity << 1);
			return;
		}

		if (existingKey3 == EMPTY) {
			keyTable[index3] = key;
			valueTable[index3] = value;
			if (size++ >= threshold) resize(capacity << 1);
			return;
		}

		push(key, value, index1, existingKey1, index2, existingKey2, index3, existingKey3);
	}

	// Skips checks for existing keys.
	private void putResize (int key, V value) {
		int[] keyTable = this.keyTable;
		int mask = this.mask;

		// Check for empty buckets.
		int index1 = key & mask;
		int existingKey1 = keyTable[index1];
		if (existingKey1 == EMPTY) {
			keyTable[index1] = key;
			valueTable[index1] = value;
			if (size++ >= threshold) resize(capacity << 1);
			return;
		}

		int index2 = hash2(key) & mask;
		int existingKey2 = keyTable[index2];
		if (existingKey2 == EMPTY) {
			keyTable[index2] = key;
			valueTable[index2] = value;
			if (size++ >= threshold) resize(capacity << 1);
			return;
		}

		int index3 = hash3(key) & mask;
		int existingKey3 = keyTable[index3];
		if (existingKey3 == EMPTY) {
			keyTable[index3] = key;
			valueTable[index3] = value;
			if (size++ >= threshold) resize(capacity << 1);
			return;
		}

		push(key, value, index1, existingKey1, index2, existingKey2, index3, existingKey3);
	}

	private void push (int key, V value, int index1, int existingKey1, int index2, int existingKey2, int index3, int existingKey3) {
		int[] keyTable = this.keyTable;
		V[] valueTable = this.valueTable;
		int mask = this.mask;

		// Push keys until an empty bucket is found.
		int insertKey = key, replacedKey;
		V insertValue = value, replacedValue;
		for (int i = 0, n = pushIterations; i < n; i++) {
			// Replace the key and value for one of the hashes.
			switch (random(3)) {
			case 0:
				replacedKey = existingKey1;
				replacedValue = valueTable[index1];
				keyTable[index1] = insertKey;
				valueTable[index1] = insertValue;
				break;
			case 1:
				replacedKey = existingKey2;
				replacedValue = valueTable[index2];
				keyTable[index2] = insertKey;
				valueTable[index2] = insertValue;
				break;
			default:
				replacedKey = existingKey3;
				replacedValue = valueTable[index3];
				keyTable[index3] = insertKey;
				valueTable[index3] = insertValue;
				break;
			}

			// If the replacedKey has an empty bucket, stop.
			index1 = replacedKey & mask;
			existingKey1 = keyTable[index1];
			if (existingKey1 == EMPTY) {
				keyTable[index1] = replacedKey;
				valueTable[index1] = replacedValue;
				if (size++ >= threshold) resize(capacity << 1);
				return;
			}

			index2 = hash2(replacedKey) & mask;
			existingKey2 = keyTable[index2];
			if (existingKey2 == EMPTY) {
				keyTable[index2] = replacedKey;
				valueTable[index2] = replacedValue;
				if (size++ >= threshold) resize(capacity << 1);
				return;
			}

			index3 = hash3(replacedKey) & mask;
			existingKey3 = keyTable[index3];
			if (existingKey3 == EMPTY) {
				keyTable[index3] = replacedKey;
				valueTable[index3] = replacedValue;
				if (size++ >= threshold) resize(capacity << 1);
				return;
			}

			insertKey = replacedKey;
			insertValue = replacedValue;
		}

		// Too many pushes occurred, increase the table size.
		resize(capacity << 1);
		put(key, value);
	}

	public V get (int key) {
		int index = key & mask;
		if (keyTable[index] == key) return valueTable[index];

		index = hash2(key) & mask;
		if (keyTable[index] == key) return valueTable[index];

		index = hash3(key) & mask;
		if (keyTable[index] == key) return valueTable[index];

		return null;
	}

	private void resize (int newSize) {
		mask = newSize - 1;
		capacity = newSize;
		threshold = (int)(newSize * loadFactor);
		hashShift = 31 - Integer.numberOfTrailingZeros(capacity);

		int[] oldKeyTable = keyTable;
		V[] oldValueTable = valueTable;

		keyTable = new int[newSize];
		int[] keyTable = this.keyTable;
		for (int i = newSize - 1; i >= 0; i--)
			keyTable[i] = EMPTY;
		valueTable = (V[])new Object[newSize];

		size = 0;
		for (int i = 0, n = oldKeyTable.length; i < n; i++) {
			int key = oldKeyTable[i];
			if (key == EMPTY) continue;
			putResize(key, oldValueTable[i]);
		}
	}

	public void clear () {
		int[] keyTable = this.keyTable;
		V[] valueTable = this.valueTable;
		for (int i = keyTable.length; i-- > 0;) {
			keyTable[i] = EMPTY;
			valueTable[i] = null;
		}
		size = 0;
	}

	private int hash2 (int h) {
		h *= PRIME2;
		return h ^ (h >>> hashShift);
	}

	private int hash3 (int h) {
		h *= PRIME3;
		return h ^ (h >>> hashShift);
	}

	public String toString () {
		if (size == 0) return "[]";
		StringBuilder buffer = new StringBuilder(32);
		buffer.append('[');

		int[] keyTable = this.keyTable;
		int i = keyTable.length;
		while (i-- > 0) {
			int key = keyTable[i];
			if (key == EMPTY) continue;
			buffer.append(key);
			buffer.append('=');
			buffer.append(valueTable[i]);
			break;
		}
		while (i-- > 0) {
			int key = keyTable[i];
			if (key == EMPTY) continue;
			buffer.append(", ");
			buffer.append(key);
			buffer.append('=');
			buffer.append(valueTable[i]);
		}
		buffer.append(']');
		return buffer.toString();
	}

	static private int randomSeed = (int)System.currentTimeMillis();

	/**
	 * Returns a random number between 0 (inclusive) and range (exclusive).
	 */
	static private final int random (int range) {
		int seed = randomSeed * 1103515245 + 12345;
		randomSeed = seed;
		return (seed >>> 15) * range >>> 17;
	}

	static private int nextPowerOfTwo (int value) {
		if (value == 0) return 1;
		if ((value & value - 1) == 0) return value;
		value |= value >> 1;
		value |= value >> 2;
		value |= value >> 4;
		value |= value >> 8;
		value |= value >> 16;
		return value + 1;
	}
}
