package fast.other;

/**
 * The ObjectStack is a stack designed specifically for storing Objects.
 * The idea is simply that it uses very little overhead and features.
 *
 * @author Joseph Lenton
 */
public class ObjectStack<A>
{
    private static final int INITIAL_SIZE = 100;
    private A[] stack;
    private int size;

    /**
     * Trivial constructor for setting up a new stack.
     */
    public ObjectStack()
    {
        this(INITIAL_SIZE);
    }

    /**
     * Creates a new stack which with space made to hold the given size.
     * @param initialSize The amount of space this stack should have by default.
     */
    public ObjectStack(int initialSize)
    {
        stack = (A[]) new Object[initialSize];
        size = 0;
    }

    /**
     * @return The number of elements stored in this stack.
     */
    public int size()
    {
        return size;
    }

    /**
     * @return True if this is empty, false if not.
     */
    public boolean isEmpty()
    {
        return size() <= 0;
    }
    
    /**
     * @param bool Adds this boolean to the top of this stack.
     */
    public void push(A bool)
    {
        ensureSize();
        stack[size++] = bool;
    }

    /**
     * @return The top element on this stack which is also permanently removed.
     */
    public A pop()
    {
        if (size <= 0) {
            throw new ArrayIndexOutOfBoundsException("Cannot pop, the stack is empty!");
        }

        A temp = stack[--size];
        stack[size] = null;
        return temp;
    }

    /**
     * Ensures that there is enough size for at least one more element to be
     * added to this stack.
     */
    private void ensureSize()
    {
        if (stack.length == size) {
            stack = ArrayUtil.copyOf(stack, (size*4) / 3);
        }
    }

    /**
     * Removes all the elements from this stack.
     */
    /* Ok, it doesn't really remove them. But if space is an issue they can use
     * compact and this is far more efficient then clearing the stack. */
    public void clear()
    {
        size = 0;
    }

    /**
     * @return The value on the top of the stack.
     */
    public A peek()
    {
        return peek( size );
    }

    /**
     * @param index The index to get the boolean from, begins at 0.
     * @return The boolean stored at the given index.
     */
    public A peek(int index)
    {
        if (index < 0 || index >= size) {
            throw new ArrayIndexOutOfBoundsException("Invalid index given: " + index);
        }

        return stack[index];
    }

    /**
     * Compacts the internal data structures down so they use as little space
     * as possible. No elements are removed by this operation.
     *
     * Ideally this should be called if the stack has had a very large number
     * of elements added, which have now been removed and will not be readded.
     */
    public void compact()
    {
        stack = ArrayUtil.copyOf(stack, size);
    }
}
